package dfki.rdf.ruleml;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import dfki.rdf.utils.*;
import dfki.rdf.ruleml.language.*;


public class XSBAgent extends HttpServlet {


  // CONFIGURE HERE
  final static String xsbProgram = "/project/java/XSB/bin/xsb";
  final static String xsbMemory = "16000"; // kb
  final static String logDir = "/home/sintek/ruleml/tomcat/logs/ruleml/";
  // constants for rdf2java:
  final static String namespace = "http://www.ruleml.org/rdf#";
  final static String pkg = "dfki.rdf.ruleml.language";
  // for transform
  final static String transformProgram = 
    "/home/sintek/ruleml/transform/transformRQ";
  // END CONFIGURE


  public void doPost(HttpServletRequest request, 
	HttpServletResponse response) throws IOException, ServletException {
    doGet(request, response);
  }

  public void doGet(HttpServletRequest request,
        HttpServletResponse response) throws IOException, ServletException {
    String query = request.getParameter("query");
    String transform = request.getParameter("transform");
    if (query != null) { // future version: different kinds of query
      query(request, response);
    } else if (transform != null) { 
      // future version: different kinds of transformations
      transform(request, response);
    } else {
      response.setContentType("text/html");
      PrintWriter htmlWriter = response.getWriter();
      htmlWriter.println("<html>");
      htmlWriter.println("<body text=#000000 bgcolor=#DCDCDC>");
      htmlWriter.println("<b>[XSBAgent] illegal request</b>");
      htmlWriter.println("</body>");
      htmlWriter.println("</html>");
    }

  }


  // query

  public void query(HttpServletRequest request,
        HttpServletResponse response) throws IOException, ServletException {

    long timestamp = new Date().getTime();
    String tmpRuleMLFile = logDir + "ruleml" + timestamp + ".rdf";
    String tmpXSBFile = logDir + "ruleml" + timestamp + ".P";

    String ruleMLProgram = request.getParameter("program");
    writeFile(tmpRuleMLFile, ruleMLProgram);

    // create html result page
    response.setContentType("text/html");
    PrintWriter htmlWriter = response.getWriter();
    htmlWriter.println("<html>");
    htmlWriter.println("<body text=#000000 bgcolor=#DCDCDC>");

    // parse rdf RuleML file and write XSB prolog file
    Map objects = null;
    try {
      objects = 
	(new RDFImport(namespace, pkg)).importObjects(tmpRuleMLFile);
      ArrayList queries = new ArrayList(); // of Query
      PrintWriter writer = new PrintWriter(new FileWriter(tmpXSBFile));
      writer.println(":- auto_table.");
      for (Iterator objIterator = objects.values().iterator(); 
	   objIterator.hasNext();) {
	Object object = objIterator.next();
	if (object instanceof Rulebase) {
	  writer.println(object);
	} else if (object instanceof Query) {
	  queries.add(object);
	}
      }
      // add queries
      for (Iterator queryIterator = queries.iterator();
	   queryIterator.hasNext();) {
	Query query = (Query)queryIterator.next();
	String name = query.getName();
	String atom = query.getAtom().toString();
	if (name == null) name = atom.toString();
	// replace ' by space; better: double ' ... !!!
	writer.println("?- writeln('% query: " + atom.replace('\'', ' ') 
	  + ".').");
	writer.println("?- " + atom + ", write_canonical(" + atom +
	  "), write('.'), nl, fail.");
      }
      writer.flush();
      writer.close();
    } catch (Exception e) {
      htmlWriter.println("[XSBAgent] error: " + e);
    }

    // execute XSB
    if (objects != null && !objects.isEmpty()) {
      String[] xsbCommand = new String[] { xsbProgram, "--noprompt", 
	"-m", xsbMemory, "-e", "['" + tmpXSBFile + "']." };
      try {
	String xsbOutput = runExternalProgram(xsbCommand, "halt.");
	htmlWriter.println("<b>Results:</b> <br>");
	htmlWriter.println("<pre>");
	htmlWriter.println(xsbOutput);
	htmlWriter.println("</pre>");
      } catch (Exception e) {
	htmlWriter.println("[XSBAgent] error: " + e);
      }
    } else { // no objects parsed!
      htmlWriter.println("<b>no rulebases or queries (parse error)</b>");
    }

    htmlWriter.println("</body>");
    htmlWriter.println("</html>");

  }


  // transform

  public void transform(HttpServletRequest request,
        HttpServletResponse response) throws IOException, ServletException {
    // response.setContentType("text/html");
    response.setContentType("text/plain");
    PrintWriter htmlWriter = response.getWriter();
    // htmlWriter.println("<html>");
    // htmlWriter.println("<body text=#000000 bgcolor=#DCDCDC>");

    long timestamp = new Date().getTime();
    String tmpPrologFile = logDir + "transform" + timestamp + ".P";
    String tmpQueriesFile = logDir + "transform" + timestamp + ".Q";

    String prologProgram = request.getParameter("program");
    writeFile(tmpPrologFile, prologProgram);
    String queries = request.getParameter("queries");
    writeFile(tmpQueriesFile, queries);

    String[] transformCommand = new String[] { transformProgram, 
      tmpPrologFile, tmpQueriesFile };
    try {
      String output = runExternalProgram(transformCommand, null);
      // htmlWriter.println("<pre>");
      htmlWriter.println(output);
      // htmlWriter.println("</pre>");
    } catch (Exception e) {
      htmlWriter.println("[XSBAgent] error: " + e);
    }

    // htmlWriter.println("</body>");
    // htmlWriter.println("</html>");
  }


  // auxiliaries 

  void writeFile(String fileName, String contents) {
    try {
      PrintWriter writer = new PrintWriter(new FileWriter(fileName));
      if (contents != null)
	writer.println(contents);
      writer.flush();
      writer.close();
    } catch (Exception e) { System.err.println(e); }
  }

  String runExternalProgram(String[] command, String sendString) 
      throws Exception {
    Process process = Runtime.getRuntime().exec(command);
    PrintWriter outputStream = new PrintWriter(process.getOutputStream(), true);
    BufferedReader reader = 
      new BufferedReader(new InputStreamReader(process.getInputStream()));
    if (sendString != null)
      outputStream.println(sendString);
    // outputStream.close(); // ???
    StringBuffer buffer = new StringBuffer();
    String line = reader.readLine();
    while (line != null) {
      buffer.append(line);
      buffer.append("\n");
      line = reader.readLine();
    }
    return buffer.toString();
  }


}


